@@ -17,10 +17,10 @@
 //! Optimizing compiler from concrete policies to Miniscript
 //!
 
-use std::collections::HashMap;
+use std::collections::BTreeMap;
 use std::convert::From;
 use std::marker::PhantomData;
-use std::{cmp, error, f64, fmt};
+use std::{cmp, error, f64, fmt, mem};
 
 use miniscript::types::extra_props::MAX_OPS_PER_SCRIPT;
 use miniscript::types::{self, ErrorKind, ExtData, Property, Type};
@@ -33,7 +33,7 @@ use {policy, Terminal};
 use {Miniscript, MiniscriptKey};
 
 type PolicyCache<Pk, Ctx> =
-    HashMap<(Concrete<Pk>, OrdF64, Option<OrdF64>), HashMap<CompilationKey, AstElemExt<Pk, Ctx>>>;
+    BTreeMap<(Concrete<Pk>, OrdF64, Option<OrdF64>), BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>>;
 
 ///Ordered f64 for comparison
 #[derive(Copy, Clone, PartialEq, PartialOrd, Debug)]
@@ -107,13 +107,13 @@ impl hash::Hash for OrdF64 {
 
 /// Compilation key: This represents the state of the best possible compilation
 /// of a given policy(implicitly keyed).
-#[derive(Copy, Clone, PartialEq, Eq, Debug, Hash)]
+#[derive(Copy, Clone, PartialEq, Eq, Debug, PartialOrd, Ord)]
 struct CompilationKey {
     /// The type of the compilation result
     ty: Type,
 
     /// Whether that result cannot be easily converted into verify form.
-    /// This is exactly the opposite of has_verify_form in the data-types.
+    /// This is exactly the opposite of has_free_verify in the data-types.
     /// This is required in cases where it is important to distinguish between
     /// two Compilation of the same-type: one of which is expensive to verify
     /// and the other is not.
@@ -161,8 +161,11 @@ struct CompilerExtData {
 
 impl Property for CompilerExtData {
     fn from_true() -> Self {
-        // only used in casts. should never be computed directly
-        unreachable!();
+        CompilerExtData {
+            branch_prob: None,
+            sat_cost: 0.0,
+            dissat_cost: None,
+        }
     }
 
     fn from_false() -> Self {
@@ -643,7 +646,7 @@ fn all_casts<Pk: MiniscriptKey, Ctx: ScriptContext>() -> [Cast<Pk, Ctx>; 10] {
 /// In general, we maintain the invariant that if anything is inserted into the
 /// map, it's cast closure must also be considered for best compilations.
 fn insert_elem<Pk: MiniscriptKey, Ctx: ScriptContext>(
-    map: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    map: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
     elem: AstElemExt<Pk, Ctx>,
     sat_prob: f64,
     dissat_prob: Option<f64>,
@@ -662,7 +665,7 @@ fn insert_elem<Pk: MiniscriptKey, Ctx: ScriptContext>(
 
     let elem_cost = elem.cost_1d(sat_prob, dissat_prob);
 
-    let elem_key = CompilationKey::from_type(elem.ms.ty, elem.ms.ext.has_verify_form, dissat_prob);
+    let elem_key = CompilationKey::from_type(elem.ms.ty, elem.ms.ext.has_free_verify, dissat_prob);
 
     // Check whether the new element is worse than any existing element. If there
     // is an element which is a subtype of the current element and has better
@@ -677,10 +680,13 @@ fn insert_elem<Pk: MiniscriptKey, Ctx: ScriptContext>(
     if !is_worse {
         // If the element is not worse any element in the map, remove elements
         // whose subtype is the current element and have worse cost.
-        map.retain(|&existing_key, existing_elem| {
-            let existing_elem_cost = existing_elem.cost_1d(sat_prob, dissat_prob);
-            !(elem_key.is_subtype(existing_key) && existing_elem_cost >= elem_cost)
-        });
+        *map = mem::replace(map, BTreeMap::new())
+            .into_iter()
+            .filter(|&(ref existing_key, ref existing_elem)| {
+                let existing_elem_cost = existing_elem.cost_1d(sat_prob, dissat_prob);
+                !(elem_key.is_subtype(*existing_key) && existing_elem_cost >= elem_cost)
+            })
+            .collect();
         map.insert(elem_key, elem);
     }
     !is_worse
@@ -695,7 +701,7 @@ fn insert_elem<Pk: MiniscriptKey, Ctx: ScriptContext>(
 /// all map is smallest possible closure of all compilations of a policy with
 /// given sat and dissat probabilities.
 fn insert_elem_closure<Pk: MiniscriptKey, Ctx: ScriptContext>(
-    map: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    map: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
     astelem_ext: AstElemExt<Pk, Ctx>,
     sat_prob: f64,
     dissat_prob: Option<f64>,
@@ -731,7 +737,7 @@ fn insert_elem_closure<Pk: MiniscriptKey, Ctx: ScriptContext>(
 fn insert_best_wrapped<Pk: MiniscriptKey, Ctx: ScriptContext>(
     policy_cache: &mut PolicyCache<Pk, Ctx>,
     policy: &Concrete<Pk>,
-    map: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    map: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
     data: AstElemExt<Pk, Ctx>,
     sat_prob: f64,
     dissat_prob: Option<f64>,
@@ -759,7 +765,7 @@ fn best_compilations<Pk, Ctx>(
     policy: &Concrete<Pk>,
     sat_prob: f64,
     dissat_prob: Option<f64>,
-) -> Result<HashMap<CompilationKey, AstElemExt<Pk, Ctx>>, CompilerError>
+) -> Result<BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>, CompilerError>
 where
     Pk: MiniscriptKey,
     Ctx: ScriptContext,
@@ -771,7 +777,7 @@ where
         return Ok(ret.clone());
     }
 
-    let mut ret = HashMap::new();
+    let mut ret = BTreeMap::new();
 
     //handy macro for good looking code
     macro_rules! insert_wrap {
@@ -811,6 +817,12 @@ where
     }
 
     match *policy {
+        Concrete::Unsatisfiable => {
+            insert_wrap!(AstElemExt::terminal(Terminal::False));
+        }
+        Concrete::Trivial => {
+            insert_wrap!(AstElemExt::terminal(Terminal::True));
+        }
         Concrete::Key(ref pk) => {
             insert_wrap!(AstElemExt::terminal(Terminal::PkH(
                 pk.to_pubkeyhash().clone()
@@ -834,11 +846,11 @@ where
             compile_binary!(&mut right, &mut left, [1.0, 1.0], Terminal::AndB);
             compile_binary!(&mut left, &mut right, [1.0, 1.0], Terminal::AndV);
             compile_binary!(&mut right, &mut left, [1.0, 1.0], Terminal::AndV);
-            let mut zero_comp = HashMap::new();
+            let mut zero_comp = BTreeMap::new();
             zero_comp.insert(
                 CompilationKey::from_type(
                     Type::from_false(),
-                    ExtData::from_false().has_verify_form,
+                    ExtData::from_false().has_free_verify,
                     dissat_prob,
                 ),
                 AstElemExt::terminal(Terminal::False),
@@ -992,9 +1004,22 @@ where
                     }
                 })
                 .collect();
+
             if key_vec.len() == subs.len() && subs.len() <= 20 {
                 insert_wrap!(AstElemExt::terminal(Terminal::Multi(k, key_vec)));
             }
+            // Not a threshold, it's always more optimal to translate it to and()s as we save the
+            // resulting threshold check (N EQUAL) in any case.
+            else if k == subs.len() {
+                let mut policy = subs.first().expect("No sub policy in thresh() ?").clone();
+                for sub in &subs[1..] {
+                    policy = Concrete::And(vec![sub.clone(), policy]);
+                }
+
+                ret = best_compilations(policy_cache, &policy, sat_prob, dissat_prob)?;
+            }
+
+            // FIXME: Should we also optimize thresh(1, subs) ?
         }
     }
     for k in ret.keys() {
@@ -1020,9 +1045,9 @@ where
 fn compile_binary<Pk, Ctx, F>(
     policy_cache: &mut PolicyCache<Pk, Ctx>,
     policy: &Concrete<Pk>,
-    ret: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
-    left_comp: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
-    right_comp: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    ret: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    left_comp: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    right_comp: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
     weights: [f64; 2],
     sat_prob: f64,
     dissat_prob: Option<f64>,
@@ -1054,10 +1079,10 @@ where
 fn compile_tern<Pk: MiniscriptKey, Ctx: ScriptContext>(
     policy_cache: &mut PolicyCache<Pk, Ctx>,
     policy: &Concrete<Pk>,
-    ret: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
-    a_comp: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
-    b_comp: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
-    c_comp: &mut HashMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    ret: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    a_comp: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    b_comp: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
+    c_comp: &mut BTreeMap<CompilationKey, AstElemExt<Pk, Ctx>>,
     weights: [f64; 2],
     sat_prob: f64,
     dissat_prob: Option<f64>,
@@ -1148,11 +1173,13 @@ mod tests {
     use super::*;
     use bitcoin::blockdata::{opcodes, script};
     use bitcoin::{self, hashes, secp256k1, SigHashType};
+    use std::collections::HashMap;
     use std::str::FromStr;
     use std::string::String;
 
     use miniscript::{satisfy, Segwitv0};
     use policy::Liftable;
+    use script_num_size;
     use BitcoinSig;
     use DummyKey;
 
@@ -1191,10 +1218,28 @@ mod tests {
         let policy = DummyPolicy::from_str(s).expect("parse");
         let miniscript: Miniscript<DummyKey, Segwitv0> = policy.compile()?;
 
-        assert_eq!(policy.lift().sorted(), miniscript.lift().sorted());
+        assert_eq!(
+            policy.lift().unwrap().sorted(),
+            miniscript.lift().unwrap().sorted()
+        );
         Ok(())
     }
 
+    #[test]
+    fn compile_timelocks() {
+        // artificially create a policy that is problematic and try to compile
+        let pol: DummyPolicy = Concrete::And(vec![
+            Concrete::Key(DummyKey),
+            Concrete::And(vec![Concrete::After(9), Concrete::After(1000_000_000)]),
+        ]);
+        assert!(pol.compile::<Segwitv0>().is_err());
+
+        // This should compile
+        let pol: DummyPolicy =
+            DummyPolicy::from_str("and(pk(),or(and(after(9),pk()),and(after(1000000000),pk())))")
+                .unwrap();
+        assert!(pol.compile::<Segwitv0>().is_ok());
+    }
     #[test]
     fn compile_basic() {
         assert!(policy_compile_lift_check("pk()").is_ok());
@@ -1231,19 +1276,25 @@ mod tests {
     fn compile_q() {
         let policy = SPolicy::from_str("or(1@and(pk(),pk()),127@pk())").expect("parsing");
         let compilation: DummySegwitAstElemExt =
-            best_t(&mut HashMap::new(), &policy, 1.0, None).unwrap();
+            best_t(&mut BTreeMap::new(), &policy, 1.0, None).unwrap();
 
         assert_eq!(compilation.cost_1d(1.0, None), 88.0 + 74.109375);
-        assert_eq!(policy.lift().sorted(), compilation.ms.lift().sorted());
+        assert_eq!(
+            policy.lift().unwrap().sorted(),
+            compilation.ms.lift().unwrap().sorted()
+        );
 
         let policy = SPolicy::from_str(
                 "and(and(and(or(127@thresh(2,pk(),pk(),thresh(2,or(127@pk(),1@pk()),after(100),or(and(pk(),after(200)),and(pk(),sha256(66687aadf862bd776c8fc18b8e9f8e20089714856ee233b3902a591d0d5f2925))),pk())),1@pk()),sha256(66687aadf862bd776c8fc18b8e9f8e20089714856ee233b3902a591d0d5f2925)),or(127@pk(),1@after(300))),or(127@after(400),pk()))"
             ).expect("parsing");
         let compilation: DummySegwitAstElemExt =
-            best_t(&mut HashMap::new(), &policy, 1.0, None).unwrap();
+            best_t(&mut BTreeMap::new(), &policy, 1.0, None).unwrap();
 
         assert_eq!(compilation.cost_1d(1.0, None), 438.0 + 299.4003295898438);
-        assert_eq!(policy.lift().sorted(), compilation.ms.lift().sorted());
+        assert_eq!(
+            policy.lift().unwrap().sorted(),
+            compilation.ms.lift().unwrap().sorted()
+        );
     }
 
     #[test]
@@ -1313,7 +1364,7 @@ mod tests {
 
         assert_eq!(ms, ms_comp_res);
 
-        let mut abs = policy.lift();
+        let mut abs = policy.lift().unwrap();
         assert_eq!(abs.n_keys(), 8);
         assert_eq!(abs.minimum_n_keys(), 2);
         abs = abs.at_age(10000);
@@ -1377,6 +1428,51 @@ mod tests {
             ]
         );
     }
+
+    #[test]
+    fn compile_thresh() {
+        let (keys, _) = pubkeys_and_a_sig(21);
+
+        // Up until 20 keys, thresh should be compiled to a multi no matter the value of k
+        for k in 1..4 {
+            let small_thresh: BPolicy = policy_str!(
+                "thresh({},pk({}),pk({}),pk({}))",
+                k,
+                keys[0],
+                keys[1],
+                keys[2]
+            );
+            let small_thresh_ms: SegwitMiniScript = small_thresh.compile().unwrap();
+            let small_thresh_ms_expected: SegwitMiniScript =
+                ms_str!("multi({},{},{},{})", k, keys[0], keys[1], keys[2]);
+            assert_eq!(small_thresh_ms, small_thresh_ms_expected);
+        }
+
+        // Above 20 keys, thresh is compiled to a combination of and()s if it's a N of N,
+        // and to a ms thresh otherwise.
+        // k = 1 (or 2) does not compile, see https://github.com/rust-bitcoin/rust-miniscript/issues/114
+        for k in &[10, 15, 21] {
+            let pubkeys: Vec<Concrete<bitcoin::PublicKey>> =
+                keys.iter().map(|pubkey| Concrete::Key(*pubkey)).collect();
+            let big_thresh = Concrete::Threshold(*k, pubkeys);
+            let big_thresh_ms: SegwitMiniScript = big_thresh.compile().unwrap();
+            if *k == 21 {
+                // N * (PUSH + pubkey + CHECKSIGVERIFY)
+                assert_eq!(big_thresh_ms.script_size(), keys.len() * (1 + 33 + 1));
+            } else {
+                // N * (PUSH + pubkey + CHECKSIG + ADD + SWAP) + N EQUAL
+                assert_eq!(
+                    big_thresh_ms.script_size(),
+                    keys.len() * (1 + 33 + 3) + script_num_size(*k) + 1 - 2 // minus one SWAP and one ADD
+                );
+                let big_thresh_ms_expected = ms_str!(
+                "thresh({},pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}),s:pk({}))",
+                k, keys[0], keys[1], keys[2], keys[3], keys[4], keys[5], keys[6], keys[7], keys[8], keys[9],keys[10], keys[11], keys[12], keys[13], keys[14], keys[15], keys[16], keys[17], keys[18], keys[19], keys[20]
+            );
+                assert_eq!(big_thresh_ms, big_thresh_ms_expected);
+            };
+        }
+    }
 }
 
 #[cfg(all(test, feature = "unstable"))]
