@@ -19,8 +19,10 @@ use bitcoin::hashes::hex::FromHex;
 use bitcoin::hashes::{hash160, ripemd160, sha256, sha256d};
 use std::{error, fmt, str};
 
+use super::ENTAILMENT_MAX_TERMINALS;
 use errstr;
 use expression::{self, FromTree};
+use miniscript::types::extra_props::{TimeLockInfo, HEIGHT_TIME_THRESHOLD};
 #[cfg(feature = "compiler")]
 use miniscript::ScriptContext;
 #[cfg(feature = "compiler")]
@@ -30,17 +32,20 @@ use policy::compiler::CompilerError;
 #[cfg(feature = "compiler")]
 use Miniscript;
 use {Error, MiniscriptKey};
-
 /// Concrete policy which corresponds directly to a Miniscript structure,
 /// and whose disjunctions are annotated with satisfaction probabilities
 /// to assist the compiler
 #[derive(Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
 pub enum Policy<Pk: MiniscriptKey> {
+    /// Unsatisfiable
+    Unsatisfiable,
+    /// Trivially satisfiable
+    Trivial,
     /// A public key which must sign to satisfy the descriptor
     Key(Pk),
-    /// A relative locktime restriction
-    After(u32),
     /// An absolute locktime restriction
+    After(u32),
+    /// A relative locktime restriction
     Older(u32),
     /// A SHA256 whose preimage must be provided to satisfy the descriptor
     Sha256(sha256::Hash),
@@ -72,6 +77,15 @@ pub enum PolicyError {
     ZeroTime,
     /// `after` fragment can only have ` n < 2^31`
     TimeTooFar,
+    /// Semantic Policy Error: `And` `Or` fragments must take args: k > 1
+    InsufficientArgsforAnd,
+    /// Semantic Policy Error: `And` `Or` fragments must take args: k > 1
+    InsufficientArgsforOr,
+    /// Entailment max terminals exceeded
+    EntailmentMaxTerminals,
+    /// lifting error: Cannot lift policies that have
+    /// a combination of height and timelocks.
+    HeightTimeLockCombination,
 }
 
 impl error::Error for PolicyError {
@@ -97,6 +111,20 @@ impl fmt::Display for PolicyError {
                 f.write_str("Relative/Absolute time must be less than 2^31; n < 2^31")
             }
             PolicyError::ZeroTime => f.write_str("Time must be greater than 0; n > 0"),
+            PolicyError::InsufficientArgsforAnd => {
+                f.write_str("Semantic Policy 'And' fragment must have atleast 2 args ")
+            }
+            PolicyError::InsufficientArgsforOr => {
+                f.write_str("Semantic Policy 'Or' fragment must have atleast 2 args ")
+            }
+            PolicyError::EntailmentMaxTerminals => write!(
+                f,
+                "Policy entailment only supports {} terminals",
+                ENTAILMENT_MAX_TERMINALS
+            ),
+            PolicyError::HeightTimeLockCombination => {
+                f.write_str("Cannot lift policies that have a heightlock and timelock combination")
+            }
         }
     }
 }
@@ -123,6 +151,8 @@ impl<Pk: MiniscriptKey> Policy<Pk> {
         Q: MiniscriptKey,
     {
         match *self {
+            Policy::Unsatisfiable => Ok(Policy::Unsatisfiable),
+            Policy::Trivial => Ok(Policy::Trivial),
             Policy::Key(ref pk) => translatefpk(pk).map(Policy::Key),
             Policy::Sha256(ref h) => Ok(Policy::Sha256(h.clone())),
             Policy::Hash256(ref h) => Ok(Policy::Hash256(h.clone())),
@@ -150,9 +180,68 @@ impl<Pk: MiniscriptKey> Policy<Pk> {
         }
     }
 
+    /// Checks whether the given concrete policy contains a combination of
+    /// timelocks and heightlocks.
+    /// Returns an error if there is atleast one satisfaction that contains
+    /// a combination of hieghtlock and timelock.
+    pub fn check_timelocks(&self) -> Result<(), PolicyError> {
+        let timelocks = self.check_timelocks_helper();
+        if timelocks.contains_combination {
+            Err(PolicyError::HeightTimeLockCombination)
+        } else {
+            Ok(())
+        }
+    }
+
+    // Checks whether the given concrete policy contains a combination of
+    // timelocks and heightlocks
+    fn check_timelocks_helper(&self) -> TimeLockInfo {
+        // timelocks[csv_h, csv_t, cltv_h, cltv_t, combination]
+        match *self {
+            Policy::Unsatisfiable
+            | Policy::Trivial
+            | Policy::Key(_)
+            | Policy::Sha256(_)
+            | Policy::Hash256(_)
+            | Policy::Ripemd160(_)
+            | Policy::Hash160(_) => TimeLockInfo::default(),
+            Policy::After(t) => TimeLockInfo {
+                csv_with_height: false,
+                csv_with_time: false,
+                cltv_with_height: t < HEIGHT_TIME_THRESHOLD,
+                cltv_with_time: t >= HEIGHT_TIME_THRESHOLD,
+                contains_combination: false,
+            },
+            Policy::Older(t) => TimeLockInfo {
+                csv_with_height: t < HEIGHT_TIME_THRESHOLD,
+                csv_with_time: t >= HEIGHT_TIME_THRESHOLD,
+                cltv_with_height: false,
+                cltv_with_time: false,
+                contains_combination: false,
+            },
+            Policy::Threshold(k, ref subs) => {
+                let iter = subs.iter().map(|sub| sub.check_timelocks_helper());
+                TimeLockInfo::combine_thresh_timelocks(k, iter)
+            }
+            Policy::And(ref subs) => {
+                let iter = subs.iter().map(|sub| sub.check_timelocks_helper());
+                TimeLockInfo::combine_thresh_timelocks(subs.len(), iter)
+            }
+            Policy::Or(ref subs) => {
+                let iter = subs
+                    .iter()
+                    .map(|&(ref _p, ref sub)| sub.check_timelocks_helper());
+                TimeLockInfo::combine_thresh_timelocks(1, iter)
+            }
+        }
+    }
+
     /// This returns whether the given policy is valid or not. It maybe possible that the policy
     /// contains Non-two argument `and`, `or` or a `0` arg thresh.
+    /// Validity condition also checks whether there is a possible satisfaction
+    /// combination of timelocks and heightlocks
     pub fn is_valid(&self) -> Result<(), PolicyError> {
+        self.check_timelocks()?;
         match *self {
             Policy::And(ref subs) => {
                 if subs.len() != 2 {
@@ -203,6 +292,7 @@ impl<Pk: MiniscriptKey> Policy<Pk> {
     ///
     pub fn is_safe_nonmalleable(&self) -> (bool, bool) {
         match *self {
+            Policy::Unsatisfiable | Policy::Trivial => (true, true),
             Policy::Key(_) => (true, true),
             Policy::Sha256(_)
             | Policy::Hash256(_)
@@ -249,6 +339,8 @@ impl<Pk: MiniscriptKey> Policy<Pk> {
 impl<Pk: MiniscriptKey> fmt::Debug for Policy<Pk> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match *self {
+            Policy::Unsatisfiable => f.write_str("UNSATISFIABLE()"),
+            Policy::Trivial => f.write_str("TRIVIAL()"),
             Policy::Key(ref pk) => write!(f, "pk({:?})", pk),
             Policy::After(n) => write!(f, "after({})", n),
             Policy::Older(n) => write!(f, "older({})", n),
@@ -290,6 +382,8 @@ impl<Pk: MiniscriptKey> fmt::Debug for Policy<Pk> {
 impl<Pk: MiniscriptKey> fmt::Display for Policy<Pk> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match *self {
+            Policy::Unsatisfiable => f.write_str("UNSATISFIABLE"),
+            Policy::Trivial => f.write_str("TRIVIAL"),
             Policy::Key(ref pk) => write!(f, "pk({})", pk),
             Policy::After(n) => write!(f, "after({})", n),
             Policy::Older(n) => write!(f, "older({})", n),
@@ -344,10 +438,14 @@ where
         }
 
         let tree = expression::Tree::from_str(s)?;
-        FromTree::from_tree(&tree)
+        let policy: Policy<Pk> = FromTree::from_tree(&tree)?;
+        policy.check_timelocks()?;
+        Ok(policy)
     }
 }
 
+serde_string_impl_pk!(Policy, "a miniscript concrete policy");
+
 impl<Pk> Policy<Pk>
 where
     Pk: MiniscriptKey,
@@ -383,6 +481,8 @@ where
             }
         }
         match (frag_name, top.args.len() as u32) {
+            ("UNSATISFIABLE", 0) => Ok(Policy::Unsatisfiable),
+            ("TRIVIAL", 0) => Ok(Policy::Trivial),
             ("pk", 1) => expression::terminal(&top.args[0], |pk| Pk::from_str(pk).map(Policy::Key)),
             ("after", 1) => {
                 let num = expression::terminal(&top.args[0], |x| expression::parse_num(x))?;
